import os
import math
import torch
import torchvision
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
import numpy as np
import matplotlib.pyplot as plt
from torchvision.datasets import ImageFolder
import torchvision.transforms as transforms
from skimage.metrics import peak_signal_noise_ratio

### Helper function(s) to use GPU instead of CPU
def get_default_device():
    """Pick GPU if available, else CPU"""
    if torch.cuda.is_available():
        return torch.device('cuda')
    else:
        return torch.device('cpu')
    
def to_device(data, device):
    """Move tensor(s) to chosen device"""
    if isinstance(data, (list,tuple)):
        return [to_device(x, device) for x in data]
    return data.to(device, non_blocking=True)

class DeviceDataLoader():
    """Wrap a dataloader to move data to a device"""
    def __init__(self, dl, device):
        self.dl = dl
        self.device = device
        
    def __iter__(self):
        """Yield a batch of data after moving it to device"""
        for b in self.dl:
            yield to_device(b, self.device)

    def __len__(self):
        """Number of batches"""
        return len(self.dl)

# Helper functions for training etc ...
@torch.no_grad()
def evaluate(model, output_activation_func, loss_func, data_loader, apply_activation_on_output_logits=True):
    # the assessment computes both accuracies and losses (on either training/validation sets)
    outputs = [model.model_assessment_step(batch, output_activation_func, loss_func, apply_activation_on_output_logits) for batch in data_loader]
    return model.model_assessment_epoch_end(outputs)

def fit(epochs, lr, model, output_activation_func, train_loader, val_loader, loss_func=nn.MSELoss(), opt_func=torch.optim.Adam, apply_activation_on_output_logits=True):
    history = []
    optimizer = opt_func(model.parameters(), lr, weight_decay=4e-6)
    optimizer = opt_func(model.parameters(), lr)
    for epoch in range(epochs):
        # Training Phase ...
        model.train() # put the model back in train phase ...
        for batch in train_loader:
            loss = model.training_step(batch, output_activation_func, loss_func, apply_activation_on_output_logits)
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
        
        # Validation phase ...
        model.eval() # put the model in the evaluation phase ...
        train_results = evaluate(model, output_activation_func, loss_func, train_loader, apply_activation_on_output_logits)
        val_results = evaluate(model, output_activation_func, loss_func, val_loader, apply_activation_on_output_logits)
        result = {'train_loss':[], 'train_psnr':[], 'val_loss':[], 'val_psnr':[]}
        result['train_loss'] = train_results["loss"]
        result['train_psnr'] = train_results["psnr"]
        result['val_loss'] = val_results["loss"]
        result['val_psnr'] = val_results["psnr"]
        model.epoch_end(epoch, result) # prints results to screen ...
        history.append(result) # compiles history for visualizing by graphs ...
    return history # returns the history [accuracies and losses] dict ...
  
# to help with plotting the psnr of our predictions...
def plot_psnr(history):
    psnrs = [x['val_psnr'] for x in history]
    plt.plot(psnrs, '-x')
    plt.xlabel('Epoch')
    plt.ylabel('PSNR')
    plt.title('PSNR vs. No. of epochs')

# to help with plotting losses during training and validation phases ...
def plot_losses(history):
    train_losses = [x.get('train_loss') for x in history]
    val_losses = [x['val_loss'] for x in history]
    plt.plot(train_losses, '-bx')
    plt.plot(val_losses, '-rx')
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.legend(['Training', 'Validation'])
    plt.title('Loss vs. No. of epochs')

# plots both accuracies and losses on the same graph ...
def plot_history(history):
  train_psnrs = [x['train_psnr'] for x in history]
  val_psnrs   = [x['val_psnr'] for x in history]
  plt.subplot(1,2,1) # make plot of accuracies ...
  plt.plot(train_psnrs, '-bx')
  plt.plot(val_psnrs, '-rx')
  plt.xlabel('Epoch')
  plt.ylabel('PSNR')
  plt.legend(['Training', 'Validation'])
  plt.title('PSNRs vs. No. of epochs')

  train_losses = [x.get('train_loss') for x in history]
  val_losses = [x['val_loss'] for x in history]
  plt.subplot(1,2,2) # make plot of losses ...
  plt.plot(train_losses, '-bx')
  plt.plot(val_losses, '-rx')
  plt.xlabel('Epoch')
  plt.ylabel('Loss')
  plt.legend(['Training', 'Validation'])
  plt.title('Loss vs. No. of epochs')

# define our custom psnr value ...
def get_psnr_value(image_true, image_test):
  err = ((image_true - image_test)**2).mean()
  signal = np.max(image_true) - np.min(image_true)
  return 10 * np.log10((signal ** 2) / err)

# we define a generic base model for ML training and evals ...
def calculate_psnr(predicted_demosaiced_images, gndtruth_images):
  psnrs = [peak_signal_noise_ratio(gndtruth_images[i,:,:,:], predicted_demosaiced_images[i,:,:,:]) for i in range(predicted_demosaiced_images.shape[0])]
  avg_psnr = np.mean(np.array(psnrs))
  return avg_psnr

class MLBaseModule(nn.Module):
    def training_step(self, batch, output_activation_func, loss_func, apply_activation_on_output_logits=True):
        raw_images, gndtruth_images = batch
        predicted_demosaiced_images = self(raw_images) # Generate predictions
        if apply_activation_on_output_logits:
          predicted_demosaiced_images = output_activation_func(predicted_demosaiced_images)
        # perform computation on gpu vs cpu
        if torch.cuda.is_available():
          gndtruth_images = gndtruth_images.to(torch.device("cuda:0")) # transport to cuda ...          
          loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.cuda.FloatTensor)) # Calculate loss
        else:
          loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.FloatTensor)) # Calculate loss
        return loss
  
    def model_assessment_step(self, batch, output_activation_func, loss_func, apply_activation_on_output_logits=True):
        raw_images, gndtruth_images = batch
        predicted_demosaiced_images = self(raw_images) # Generate predictions
        if apply_activation_on_output_logits:
          predicted_demosaiced_images = output_activation_func(predicted_demosaiced_images)
        # perform computation on gpu vs cpu
        if torch.cuda.is_available():
          gndtruth_images = gndtruth_images.to(torch.device("cuda:0")) # transport to cuda ...          
          loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.cuda.FloatTensor)) # Calculate loss
        else:
          loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.FloatTensor)) # Calculate loss
        batch_psnr = calculate_psnr(predicted_demosaiced_images.detach().cpu().numpy(), gndtruth_images.detach().cpu().numpy())           # Calculate psnr
        return {'loss': loss.detach(), 'psnr': batch_psnr}
        
    def model_assessment_epoch_end(self, outputs):
        batch_losses = [x['loss'] for x in outputs]
        epoch_loss = np.array(batch_losses).mean()   # Combine losses
        batch_psnrs = [x['psnr'] for x in outputs]
        epoch_psnr = np.array(batch_psnrs).mean()    # Combine accuracies
        return {'loss': epoch_loss, 'psnr': epoch_psnr}
    
    def epoch_end(self, epoch, result):
        print("Epoch {}: train_loss: {:.4f}, val_loss: {:.4f}, train_psnr: {:.4f}, val_psnr: {:.4f}".format(
            epoch+1, result['train_loss'], result['val_loss'], result['train_psnr'], result['val_psnr']))

class MLEncoderBaseModule(nn.Module): 
    def training_step(self, batch, output_activation_func, loss_func, apply_activation_on_output_logits=True):
        raw_images, gndtruth_images = batch
        predicted_raw_images, predicted_demosaiced_images = self(raw_images) # Generate predictions
        if apply_activation_on_output_logits:
          predicted_raw_images = output_activation_func(predicted_raw_images)
          # predicted_demosaiced_images = output_activation_func(predicted_demosaiced_images)
        # perform computation on gpu vs cpu
        if torch.cuda.is_available():
          # gndtruth_images = gndtruth_images.to(torch.device("cuda:0")) # transport to cuda ...          
          raw_images = raw_images.to(torch.device("cuda:0")) # transport to cuda ...          
          # loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.cuda.FloatTensor)) # Calculate loss
          loss = loss_func(predicted_raw_images, raw_images.type(torch.cuda.FloatTensor)) # Calculate loss
        else:
          # loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.FloatTensor)) # Calculate loss
          loss = loss_func(predicted_raw_images, raw_images.type(torch.FloatTensor)) # Calculate loss
        return loss
  
    def model_assessment_step(self, batch, output_activation_func, loss_func, apply_activation_on_output_logits=True):
        raw_images, gndtruth_images = batch
        predicted_raw_images, predicted_demosaiced_images = self(raw_images) # Generate predictions
        if apply_activation_on_output_logits:
          predicted_raw_images = output_activation_func(predicted_raw_images)
          # predicted_demosaiced_images = output_activation_func(predicted_demosaiced_images)
        # perform computation on gpu vs cpu
        if torch.cuda.is_available():
          # gndtruth_images = gndtruth_images.to(torch.device("cuda:0")) # transport to cuda ...          
          raw_images = raw_images.to(torch.device("cuda:0")) # transport to cuda ...          
          # loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.cuda.FloatTensor)) # Calculate loss
          loss = loss_func(predicted_raw_images, raw_images.type(torch.cuda.FloatTensor)) # Calculate loss
        else:
          # loss = loss_func(predicted_demosaiced_images, gndtruth_images.type(torch.FloatTensor)) # Calculate loss
          loss = loss_func(predicted_raw_images, raw_images.type(torch.FloatTensor)) # Calculate loss
        batch_psnr = calculate_psnr(predicted_raw_images.detach().cpu().numpy(), raw_images.detach().cpu().numpy())           # Calculate psnr
        return {'loss': loss.detach(), 'psnr': batch_psnr} # we compute the psnr of only the reconstructed colored images ...
        
    def model_assessment_epoch_end(self, outputs):
        batch_losses = [x['loss'] for x in outputs]
        epoch_loss = np.array(batch_losses).mean()   # Combine losses
        batch_psnrs = [x['psnr'] for x in outputs]
        epoch_psnr = np.array(batch_psnrs).mean()    # Combine accuracies
        return {'loss': epoch_loss, 'psnr': epoch_psnr}
    
    def epoch_end(self, epoch, result):
        print("Epoch {}: train_loss: {:.4f}, val_loss: {:.4f}, train_psnr: {:.4f}, val_psnr: {:.4f}".format(
            epoch+1, result['train_loss'], result['val_loss'], result['train_psnr'], result['val_psnr']))